#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "conway.h"
#include "cuda.h"

#define N					128
#define MAXITERS			10000
#define THREADS_PER_BLOCK	16

__device__ int count_neighbors_gpu(char* b, int n, int m, int i, int j)
{
	int di, dj, im, jm;
	int sum = 0;

	for (di = -1; di < 2; di++)
		for (dj = -1; dj < 2; dj++)
			if (di != 0 || dj != 0)
			{
				im = (i + di + n) % n;
				jm = (j + dj + m) % m;
				sum = sum + b[im * m + jm];
			}

	return sum;
}

__global__ void ConwaysCell(char *board_new, char *board, int height, int width)
{
    __shared__ char loc[(THREADS_PER_BLOCK+2)*(THREADS_PER_BLOCK+2)];

	int gi = blockDim.y * blockIdx.y + threadIdx.y;
	int gj = blockDim.x * blockIdx.x + threadIdx.x;
	int li = threadIdx.y;
	int lj = threadIdx.x;
	int lk = blockDim.x * li + lj;			// 2D -> 1D local index
	int oi = blockDim.y * blockIdx.y - 1;	// origin @ board
    int oj = blockDim.x * blockIdx.x - 1;
	int bi, bj;

    while (lk < (THREADS_PER_BLOCK+2)*(THREADS_PER_BLOCK+2))
    {
        bi = (oi + lk / (blockDim.x+2) + height) % height;
        bj = (oj + lk % (blockDim.x+2) + width) % width;
        loc[lk] = board[width*bi+bj];
        lk += blockDim.y * blockDim.x;
    }

	__syncthreads();

	if (gi < height && gj < width)
	{
		int neighs = count_neighbors_gpu(loc, blockDim.y+2, blockDim.x+2, li+1, lj+1);

		if (neighs == 3 || (loc[(li+1)*(blockDim.x+2)+(lj+1)] == 1 && neighs == 2))
			board_new[width*gi+gj] = 1;
		else
			board_new[width*gi+gj] = 0;
	}
}

int main(int argc, char *argv[])
{
	// initialize board
	char **board, **board_new;
	srand(1573949136);
	board = board_initialize(N, N);
	board_new = board_initialize(N, N);
	board_print(board, N, N);

	// allocate device memory
	char *board_d, *board_new_d;
    cudaMalloc((void **)&board_d, N * N * sizeof(char));
    cudaMalloc((void **)&board_new_d, N * N * sizeof(char));

    // copy data to device
    cudaMemcpy(board_d, board[0], N * N * sizeof(char), cudaMemcpyHostToDevice);

	// divide work among threads
    dim3 threadsperblock(THREADS_PER_BLOCK, THREADS_PER_BLOCK);
    dim3 blockspergrid((N-1)/THREADS_PER_BLOCK + 1, (N-1)/THREADS_PER_BLOCK + 1);

	// timing - start
	float elapsedTime;
    cudaEvent_t start, stop;
    cudaEventCreate(&start);
    cudaEventCreate(&stop);
    cudaEventRecord(start);

	// evolve
	int iters = 0;
	while (iters < MAXITERS)
	{
		// trigger kernel execution
		if (iters % 2 == 0)
			ConwaysCell<<<blockspergrid, threadsperblock>>>(board_new_d, board_d, N, N);
		else
			ConwaysCell<<<blockspergrid, threadsperblock>>>(board_d, board_new_d, N, N);

		cudaError_t err = cudaGetLastError();
		if (err != cudaSuccess) 
		{
			printf("cudaGetDeviceCount error %d\n-> %s\n", err, cudaGetErrorString(err));
			exit(EXIT_FAILURE);
		}
		iters++;		
	}

	// timing - stop
    cudaEventRecord(stop);
    cudaEventSynchronize(stop);
    cudaEventElapsedTime(&elapsedTime, start, stop);

    // copy data to host
    cudaMemcpy(board[0], iters % 2 == 0 ? board_d : board_new_d, N * N * sizeof(char), cudaMemcpyDeviceToHost);

	// print the result
	board_print(board, N, N);
	printf("\n\nTime: %lf ms\n\n", elapsedTime);

	// free device memory
	cudaFree(board_d);
	cudaFree(board_new_d);

	// free host memory
	board_free(board);
	board_free(board_new);

	return 0;
}
